from typing import Tuple
import reg
import cfl
import sys
import signal


def get_task(string: str) -> Tuple[str, str]:
    assert '-' in string, "Teacher solution could not be parsed correctly."
    teacher_type, task = string.split("-", 1)
    if '-' in task:
        task, _ = task.split("-", 1)
    assert len(teacher_type) != 0, "Teacher solution code must not ne empty"
    assert len(task) != 0, "Task code must not be empty"

    return (teacher_type, task)


def dfa_transform(string: str, automaton_type: str) -> reg.DFA:
    try:
        parser = reg.Parser()
        if automaton_type in {"DFA", "TOT", "MIN", "TOC", "MIC"}:
            automaton = parser.str_to_dfa(string)
        elif automaton_type == "NFA":
            automaton = parser.str_to_nfa(string).determinize()
        elif automaton_type == "EFA":
            automaton = parser.str_to_nfa(string).eliminate_epsilon().determinize()
        elif automaton_type == "GRA":
            automaton = parser.str_to_reggrammar(string).reggrammar_to_nfa().determinize()
        elif automaton_type == "REG":
            automaton = parser.str_to_regex(string).regex_to_efa().eliminate_epsilon().determinize()

        return automaton

    # TODO K: this does nothing yet!
    except reg.ParsingError as ex:
        print("Chyba při parsování.")
        exit(1)


def nfa_transform(string: str, automaton_type: str) -> reg.NFA:
    try:
        parser = reg.Parser()
        return parser.str_to_nfa(string)

    # TODO K: errors cleanup
    except reg.ParsingError as message:
        print("Parsing error:", message)
        exit(1)


def check_task(dfa: reg.DFA, task: str) -> bool:
    if task not in {"TOT", "MIN", "TOC", "MIC"}:
        return True

    if task == "TOT" and not dfa.is_total():
        print("DFA není totální.")
        return False

    if task == "MIN" and not dfa.is_minimal():
        print("DFA není minimální.")
        return False

    canonical = dfa.is_canonical()
    if task == "CAN" and not canonical:
        print("DFA není kanonický.")
        return False

    if task == "TOC" and (not dfa.is_total() or not canonical):
        if not dfa.is_total():
            print("DFA není totální.")
        if not canonical:
            print("DFA není kanonický.")
        return False

    if task == "MIC" and (not dfa.is_minimal() or not canonical):
        if not dfa.is_minimal():
            print("DFA není minimální.")
        if not canonical:
            print("DFA není kanonický.")
        return False

    return True


def check_empty(student_solution, teacher_solution):
    if teacher_solution.is_empty() and student_solution.is_empty():
        exit_correct()


def check_alphabets(student_alpha, teacher_alpha, task: str) -> None:
    if student_alpha != teacher_alpha:
        print("Abecedy zadaného jazyka a studentova řešení se liší")

        if task == "REG":
            print("(abeceda je odvozena ze všech písmen objevujících se v regulárním výrazu).")
        elif task == "GRA":
            print("(abeceda je odvozena ze všech terminálů objevujících se v pravidlech gramatiky).")
        else:
            print("(abeceda je odvozena ze všech písmen objevujících se v přechodové funkci).")

        # TODO ask also print both alphabets? Or difference?
        #print("Studentova abeceda:" + reg.Parser.names_to_str(student_alpha))
        #print("Abeceda zadaného jazyka:" + reg.Parser.names_to_str(teacher_alpha))

        exit_incorrect()


def exit_correct():
    print("Správné řešení.")
    exit(0)


def exit_incorrect():
    print("Nesprávné řešení.")
    exit(1)


def print_extra_word_ce(student_word):
    print("Příklad slova, které je ve studentově řešení a není v zadaném jazyce: ", student_word)


def print_missing_word_ce(teacher_word):
    print("Příklad slova, které chybí ve studentově řešení a je v zadaném jazyce: ", teacher_word)


def dfa_task(teacher_type, teacher_string, task, student_string):
    student_solution = dfa_transform(student_string, task)
    teacher_solution = dfa_transform(teacher_string, teacher_type)

    check_empty(student_solution=student_solution,
                teacher_solution=teacher_solution)

    check_alphabets(student_alpha=student_solution.characters,
                    teacher_alpha=teacher_solution.characters, task=task)

    result = reg.DFA.is_equivalent(student_solution, teacher_solution)
    if check_task(student_solution, task) and result:
        exit_correct()

    if not result:
        student_word, teacher_word = result.left_counterexample, result.right_counterexample

        if student_word is None:
            print("Studentovo řešení je podmnožinou zadaného jazyka.")
        else:
            print_extra_word_ce(student_word)

        if teacher_word is None:
            print("Studentovo řešení je nadmnožinou zadaného jazyka.")
        else:
            print_missing_word_ce(teacher_word)

        if result.inf is not None:
            print("Rozdíl porovnávaných jazyků je nekonečný.")
        else:
            print("Rozdíl porovnávaných jazyků je konečný.")

    exit_incorrect()


def exit_cfl_ok_but_invalid_constraint(msg : str) -> None:
    print("Gramatika generuje zadaný jazyk, ale nesplňuje podmínky "
          f"zadání: {msg}")
    exit(1)


def cfg_task(teacher_type: str, teacher_string: str, task: str,
             student_string: str) -> None:
    parser = reg.Parser()
    assert task[:3] == "CFG", f"Invalid task prefix {task[:3]}"
    if len(task) > 3:
        assert task[3] == '+', f"Invalid task suffix {task[3:]}"
    constraints = filter(lambda x: len(x) != 0, task[4:].split(","))
    try:
        teacher_solution = parser.str_to_cfg(teacher_string)
    except reg.ParsingError as message:
        print(f"Error parsing teacher's solution: {message}")
        exit(1)

    try:
        student_solution = parser.str_to_cfg(student_string)
    except reg.ParsingError as message:
        print(f"Error parsing student's solution: {message}")
        exit(1)

    check_empty(student_solution=student_solution,
                teacher_solution=teacher_solution)

    check_alphabets(student_alpha=student_solution.terminals,
                    teacher_alpha=teacher_solution.terminals, task="GRA")

    equals = cfl.CFG.is_equivalent_test(student_solution, teacher_solution)

    if equals:
        failed = []
        for constraint in constraints:
            if constraint == "¬ε":
                if not student_solution.is_epsilon_normal_form():
                    failed.append("obsahuje nepovolené ε-kroky")
            elif constraint == "¬s":
                if student_solution.has_simple_rules():
                    failed.append("obsahuje nepovolená jednoduchá pravidla")
            elif constraint == "CNF":
                if not student_solution.is_cnf():
                    failed.append("není v CNF")
            else:
                assert False, f"Unknown constraint `{constraint}'"

        if len(failed) == 0:
            exit_correct()
        exit_cfl_ok_but_invalid_constraint(", ".join(failed))

    if equals.left_counterexample is not None:
        print_extra_word_ce(equals.left_counterexample)
    if equals.right_counterexample:
        print_missing_word_ce(equals.right_counterexample)

    exit_incorrect()


def main():
    signal.alarm(50)

    # Get string with student's solution.
    with open(sys.argv[1]) as student_file:
        student_string = student_file.read()

    # Get string with teacher's solution and type of the task.
    for argument in sys.argv:
        if argument[:2] == "-o":
            teacher_string = argument[2:]

    try:
        task_prefix, teacher_string = teacher_string.split(":", 1)
        teacher_type, task = get_task(task_prefix)

        if teacher_type == "DFA" or teacher_type == "NFA":
            dfa_task(teacher_type=teacher_type, teacher_string=teacher_string,
                     task=task, student_string=student_string)
        elif teacher_type == "CFG":
            cfg_task(teacher_type=teacher_type, teacher_string=teacher_string,
                     task=task, student_string=student_string)
        else:
            print(f"Invalid question type {task_prefix}")
            exit(1)

    # TODO K: get your exceptions together!
    except Exception as ex:
    #except AttributeError as message:
        print("Error inside of fja_checker:", ex.args)
        exit(1)

if __name__ == "__main__":
    main()
